/*
 * Copyright (c) y2k Micha Moskal <malekith@topnet.pl>
 *
 * This file is part of cryptoFTP - cftp protocol implementation.
 * CryptoFTP is free software. You can redistribute it and/or modify it under
 * the terms of the GNU General Public License (see COPYING.GPL for details).
 *
 * $Id: transfer.c,v 1.2 2000/04/22 16:19:40 malekith Exp $
 */

#include "config.h"
#include "cftp.h"
#include "utils.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h> 

extern int abort_please;
extern struct clib_state *ctx;

static void print_bytes(double bytes)
{
	if (bytes < 1024)
		printf("%7d B", 
		       (int)bytes);
	else if (bytes < 100 * 1024) 
		printf("%2d.%03d kB", 
		       (int)bytes/1024, (int)(bytes/1.024)%1000);
	else if (bytes < 1024 * 1024)
		printf("%4d.%01d kB", 
		       (int)bytes/1024,(int)(bytes/102.4)%10);
	else
		printf("%3d.%02d MB", 
		       (int)bytes/(1024*1024), (int)(bytes/(1024*10.24))%100);
}

#ifdef USE_PBAR
/*
 * this has to made cursor go one line up
 */
static void go_up()
{
	printf("\033M");
}
#endif /* USE_PBAR */

static void display_pbar(int bytes, int size)
{
	static double last_time, start_time;
	static int last_bytes;
	static double trans, last_display;
	double t;
	int t1, t2;

	if (bytes == 0) {
		start_time = last_time = dtime();
		last_bytes = 0;
		trans = 0;
		printf("\n");
	}
	
	if (dtime() - last_display < 0.1 && bytes != size)
		return;
	last_display = dtime();
	
	if (dtime() - last_time > 1) {
		trans = (bytes - last_bytes) / (dtime() - last_time);
		last_time = dtime();
		last_bytes = bytes;
	}

	t = dtime() - start_time;
	
	t1 = (int)t;
	t2 = (size && bytes ? (int)(t * size / bytes) : 0);
	
#ifdef USE_PBAR
	go_up();
#endif /* USE_PBAR */

	printf("\r-> ");
	print_bytes(bytes); printf(" / ");
	print_bytes(size);
	printf("  (%3d%%) %3d:%02d/%3d:%02d at ", 
	       (size ? bytes * 100 / size : 0), t1 / 60, 
	       t1 % 60, t2 / 60, t2 % 60);
	print_bytes(trans);
	printf("/s  ");
	
#ifdef USE_DASH
	{
		int zzz = bytes + 1000;
		const char *dash = "----=[|]=-";
		int k, n;
		
		n = k = strlen(dash);
		
		if (trans > 200000)
			zzz /= (1000000/k);
		else if (trans > 10000)
			zzz /= (2000/k);
		else
			zzz /= (200/k);

		while (n-->0)
			printf("%c", dash[(zzz-n)%k]);
	}
#endif

	printf("  <-\r");
	
#ifdef USE_PBAR
	{
	int n;
#define LEN 75
	printf("\n[");
	n = LEN;
	while (n-- > 0) {
		int d = ((double)bytes / size - (double)(LEN-n) / LEN) * 100;
#ifdef USE_FANCY_PBAR
		const char *k = " >-=*%#";
		d += 2;
		if (d < 0) d = 0;
		if (d > 6) d = 6;
		printf("%c",k[d]);
#else /* USE_FANCY_PBAR */
		printf("%c", d >= 0 ? '#' : ' ');
#endif /* USE_FANCY_PBAR */
	}
	printf("]\r");
#undef LEN
	}
#endif /* USE_PBAR */

	fflush(stdout);
}

int z_decompress_size(struct clib_state *ctx);

static void do_get_file(int size, FILE * out)
{
	int n = 0, ds;
	int aborting = 0;
	double start_time;
	int err;

	start_time = dtime();

	display_pbar(0, size);
	abort_please = 0;

	for (;;) {
		void *buf;
		int bufsize;

		err = clib_read(ctx, &buf, &bufsize);
		n += bufsize;
		fwrite(buf, 1, bufsize, out);
		display_pbar(n, size);

		xfree(buf);

		if (err)
			break;

		if (abort_please && !aborting) {
			aborting = 1;
			clib_abort_get(ctx);
			printf("\nabort request sent\n\n");
		}
	}

	start_time = (dtime() - start_time) * 1024;
	ds = z_decompress_size(ctx);
	if (start_time != 0 && ds)
		printf("\n%%%%%% real bytes transfered over network: %d "
		       "[%.2f kB/s]", ds, ds / start_time);
	else if (start_time != 0)
		printf("\n%%%%%% overall transfer: %.2f kB/s",
		       n / start_time);

	printf("\n*** transfer %s\n",
	       err == 1 ? "OK" : (err == 2 ? "ABORTED" : ""));
}

void get_file(const char *name, const char *save_as, int compress)
{
	int mode, size;
	char *owner, *group;
	FILE *out;

	out = fopen(save_as, "wb");
	if (!out) {
		fprintf(stderr, "!> couldn't open local file ");
		perror(save_as);
		return;
	}

	if (clib_get(ctx, name, 1, compress, 0, 0, &mode, &size,
		     &owner, &group)) {
		fclose(out);
		return;
	}
	if (strcmp(name, save_as) == 0)
		printf("getting: '%s' (mode: 0%o, owner: %s.%s)", name, mode,
		     owner, group);
	else
		printf("getting: '%s' (mode: 0%o, owner: %s.%s) --> %s",
		     name, mode, owner, group, save_as);
	mode &= 01777;		/* don't transfer suid bits */

	do_get_file(size, out);
	fclose(out);
	out = 0;
	chmod(save_as, mode);
}

static void do_send_file(FILE * f, int size)
{
	int n = 0;
	int err;

	display_pbar(0, size);
	abort_please = 0;

	for (n = 0; n < size && !abort_please;) {
		int r;
		byte buf[CHUNK_SIZE];

		r = CHUNK_SIZE;
		if (r > size - n)
			r = size - n;

		r = fread(buf, 1, r, f);
		n += r;
		if (clib_write(ctx, buf, r))
			break;
		display_pbar(n, size);
	}
	err = clib_stor_finish(ctx);

	printf("\n*** transfer %s\n",
	       err == 0 ? "OK" : 
		       	((err == 1 && abort_please) ? "ABORTED" : "ERROR"));
}


void send_file(const char *name, const char *send_as, int compress)
{
	struct stat st;
	FILE *in;

	in = fopen(name, "rb");
	if (!in) {
		printf("!> couln't read local file ");
	      perror(name);
		return;
	}

	stat(name, &st);

	if (clib_stor(ctx, send_as, CLIB_STOR_TRUNC, st.st_mode & 07777, 0,
		      st.st_size, compress)) {
		fclose(in);
		return;
	}

	if (strcmp(name, send_as) == 0)
		printf("sending: '%s' (mode: 0%o)", name,
		     st.st_mode & 07777);
	else
		printf("sending: '%s' (mode: 0%o) --> %s", name,
		     st.st_mode & 07777, send_as);

	do_send_file(in, st.st_size);
}
