/*
 * Copyright (c) y2k Micha Moskal <malekith@topnet.pl>
 *
 * This file is part of cryptoFTP - cftp protocol implementation.
 * CryptoFTP is free software. You can redistribute it and/or modify it under
 * the terms of the GNU General Public License (see COPYING.GPL for details).
 *
 * $Id: forward.c,v 1.7 2000/05/05 22:20:26 malekith Exp $
 */

#include "config.h"
#include "server.h"

#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/time.h>
#include <ctype.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>

extern struct clib_state *ctx;

void key_rsa();

static int create_connection(const char *host, int port)
{
	struct sockaddr_in sin;
#ifdef IPv6
	struct sockaddr_in6 sin6;
	int use6 = 0;
#endif
	int sock, r;
	
#ifndef IPv6
	/* in case of ipv6 we can't simply assume that address starting with
	 * digit is ipv4... gethostbyname should return proper info for
	 * both dot-quad and fqdn notation... */
	if (isdigit(*host))
		*(u32 *) & sin.sin_addr = inet_addr(host);
	else 
#endif
	{
		struct hostent *h;
		
		h = gethostbyname(host);
		if (!h) die("unknown host : %s", host);
#ifdef IPv6
		if (h->h_addrtype == AF_INET6) {
			memcpy(sin6.sin6_addr.s6_addr, h->h_addr_list[0], 16);
			use6 = 1;
		} else
#endif
			memcpy(&sin.sin_addr, h->h_addr_list[0], 4);
	}
	
#ifdef IPv6
	if ((sock = socket(use6 ? PF_INET6 : PF_INET, SOCK_STREAM, 0)) == -1)
#else
	if ((sock = socket(PF_INET, SOCK_STREAM, 0)) == -1)
#endif
		/* it's possibly fatal -- NFILES limit reached or something */
		die("socket"); 
		
#ifdef IPv6
	if (use6) {
		sin6.sin6_family = AF_INET6;
		sin6.sin6_port = htons(port);
		r = connect(sock, (struct sockaddr*)&sin6, sizeof(sin6));
	} else
#endif
	{
		sin.sin_family = AF_INET;
		sin.sin_port = htons(port);
		r = connect(sock, (struct sockaddr*)&sin, sizeof(sin));
	}
	
	if (r == -1)
		die("couldn't connect to : %s", host);

	return sock;		
}

static void fwd_loop(int port, const char *host)
{
	int sock = -1;

	log_init();
	info("forwarding server startup");
	
	for (;;) {
		fd_set rd;
		int maxsock;

		FD_ZERO(&rd);
		FD_SET(ctx->sock, &rd);
		if (sock != -1)
			FD_SET(sock, &rd);
		maxsock = sock;
		if (maxsock == -1 || ctx->sock > maxsock)
			maxsock = ctx->sock;
			
		if (select(maxsock + 1, &rd, 0, 0, 0) < 0)
			die("select() error");
		
		if (FD_ISSET(ctx->sock, &rd)) {
			struct packet *p;

			p = recv_packet(ctx);
			if (!p)
				return;
			switch (p->id) {
			case PKT_DISCONNECT:
				info("server shutdown");
				exit(0);
			case PKT_COOKIE:
				send_packet(ctx, p->id, p->data, p->size);
				break;
			case PKT_FORWARD_DATA:
				if (sock != -1)
					write_l(sock, p->data, p->size);
				break;
			case PKT_FORWARD_CONNECT:
				info("got fwd connect");
				if (sock != -1)
					close(sock);
				sock = create_connection(host, port);
				break;
			case PKT_FORWARD_EOF:
				info("got fwd eof");
				close(sock);
				sock = -1;
				break;
			default:
				die("got unknown packet: %d", p->id);
			}
			free_packet(p);
		}
		if (sock != -1 && FD_ISSET(sock, &rd)) {
			byte buf[CHUNK_SIZE];
			int r;

			r = read(sock, buf, sizeof(buf));
			if (r <= 0) {
				send_packet(ctx, PKT_FORWARD_EOF, "", 0);
				close(sock);
				sock = -1;
			} else
				send_packet(ctx, PKT_FORWARD_DATA, buf, r);
		}
	}
}

extern struct cipher2 *using_cipher;
extern int using_key_len, shell_type;
extern char *user_name;

#ifdef IPv6
extern int use_ipv6;
#endif

void xforward(struct pkt_command *c)
{
	struct cmd_forward *f;
	struct pkt_response *r;
	struct rsp_forward *rf;
	const char *host = "localhost";
	int sock, n, pid, port;
	struct sockaddr_in sin;
#ifdef IPv6
	struct sockaddr_in6 sin6;
#endif

	socklen_t len;

	f = (struct cmd_forward *) c->data;
	f->port = n2h(f->port);

	if (user_name) {
		if (conf_get_bool("AllowNonLocalForwards", 0)
		    && f->host[0]) host = f->host;
		else if (f->host[0] && strcmp("localhost", f->host) != 0)
			string_resp(RSP_STATUS_ERR,
				    "can't forward to non-local hosts");

		if (conf_get_bool("ForwardRequiresValidShell", 0) &&
		    (shell_type & 1))
			string_resp(RSP_STATUS_ERR,
				    "valid shell is required "
				    "to request forwarding -- contact sysadmin "
				    "to get one");


		if (conf_get_bool("ForwardRequiresNonFakeShell", 0) &&
		    (shell_type & 2))
			string_resp(RSP_STATUS_ERR, "non-fake shell is "
				    "required to request forwarding -- contact "
				    "sysadmin o get one");

		info("FORWARD %s @ %d ok", host, f->port);
	} else {
		char *target;
		int tmp;

		/* just in case */
		setuid(((tmp = conf_get_int("NobodyUID", 99))) ? tmp : 99);
		setgid(((tmp = conf_get_int("NobodyGID", 99))) ? tmp : 99);

		if (f->host[0])
			host = f->host;

		target = xprintf("%s@%d", host, f->port);

		if (conf_find_string("AnonForward", target) == 0) {
			warn("not allowed forward reqested: %s", target);
			string_resp(RSP_STATUS_ERR, "sorry....");
			exit(0);
		}
		notice("anon forward %s", target);
	}

	host = xstrdup(host);

#ifdef IPv6
	if (use_ipv6) {
		sock = socket(PF_INET6, SOCK_STREAM, 0);

		sin6.sin6_family = AF_INET6;
		sin6.sin6_addr = in6addr_any;
		sin6.sin6_port = htons(0);

		if (bind(sock, (struct sockaddr *) &sin6, sizeof(sin6)) <
		    0) die("bind6");

		len = sizeof(sin6);
		if (getsockname(sock, (struct sockaddr *) &sin6, &len) < 0)
			die("getsockname6");
		port = ntohs(sin6.sin6_port);
	} else
#endif
	{
		sock = socket(PF_INET, SOCK_STREAM, 0);

		sin.sin_family = AF_INET;
		sin.sin_addr.s_addr = htonl(INADDR_ANY);
		sin.sin_port = htons(0);

		if (bind(sock, (struct sockaddr *) &sin, sizeof(sin)) < 0)
			die("bind");

		len = sizeof(sin);
		if (getsockname(sock, (struct sockaddr *) &sin, &len) < 0)
			die("getsockname");
		port = ntohs(sin.sin_port);
	}

	/* n = sizeof(struct pkt_response) + sizeof(struct rsp_forward) - 1; */
	n = 4 + 8;
	r = (struct pkt_response *) xmalloc(n);
	r->no = h2n(RSP_FORWARD);
	rf = (struct rsp_forward *) r->data;
	rf->flags = h2n(0);
	rf->port = h2n(port);
	debug("fwd port %d",port);
	send_packet(ctx, PKT_RESPONSE, (byte *) r, n);
	xfree0(r, n);
	pid = fork();

	if (pid == -1)
		die("fork");

	if (pid) {
		close(sock);
		if (user_name)
			return;
		else
			exit(0);
	}
        
	debug("after fork");

	listen(sock, 1);
        
debug("after listen");
/*
len=sizeof(sin);
{
fd_set rd;FD_ZERO(&rd); FD_SET(sock,&rd);
select(sock+1,&rd,0,0,0);
}
	n = accept(sock, &sin, &len);*/
        n = accept(sock, 0, 0);
	if (n < 0)
		die("accept");
	close(sock);
	sock = n;
        debug("got client");
        
#ifndef BRAINDEAD
	closelog();
	n = 512;
	while (n-- >= 0)
		if (n != sock)
			close(n);
#endif
	ctx->sock = sock;

	/* setup new keys */
	ctx->u->e->free(ctx->u->e);
	ctx->u->d->free(ctx->u->d);
	ctx->u->e = using_cipher->creator();
	ctx->u->d = using_cipher->creator();
	ctx->u->e->init(ctx->u->e, f->server_key, using_key_len);
	ctx->u->d->init(ctx->u->d, f->client_key, using_key_len);
	ctx->u->encrypting = 1;

	memset(f->server_key, 0, 32);
	memset(f->client_key, 0, 32);
	fwd_loop(f->port, host);
	exit(1);
}
