/*
 * Copyright (c) y2k Micha Moskal <malekith@topnet.pl>
 *
 * This file is part of cryptoFTP - cftp protocol implementation.
 * CryptoFTP is free software. You can redistribute it and/or modify it under
 * the terms of the GNU General Public License (see COPYING.GPL for details).
 *
 * $Id: log.c,v 1.5 2000/05/04 19:09:35 malekith Exp $
 */

#include "config.h"
#include "server.h"

#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>

#ifndef BRAINDEAD
#include <syslog.h>
#else
static void syslog(int dummy, const char *fmt, ... );
#define LOG_DEBUG 0
#define LOG_INFO 0
#define LOG_NOTICE 0
#define LOG_WARNING 0
#define LOG_ERR 0
#define LOG_CRIT 0
#define LOG_DAEMON 0
#define LOG_FTP 0
#define LOG_USER 0
#define LOG_LOCAL0 0
#define LOG_LOCAL1 0
#define LOG_LOCAL2 0
#define LOG_LOCAL3 0
#define LOG_LOCAL4 0
#define LOG_LOCAL5 0
#define LOG_LOCAL6 0
#define LOG_LOCAL7 0
#endif

/* default values */
static int log_level = 2;	/* notice */
static int syslog_f = LOG_DAEMON;
extern char *user_name;

#ifndef LOG_FTP
#define LOG_FTP (11<<3)
#endif
 
char *xvprintf(const char *fmt, va_list ap);

static void log(int severity, const char *fmt, va_list ap)
{
	char *p;
	const char *un;
	int sf = syslog_f;
	
	if ((un = user_name) == 0)
		un = "N0NE";
		
	if (severity < log_level)
		return;
		
	p = xvprintf(fmt, ap);
	
	switch (severity) {
	case 0:
		syslog(sf|LOG_DEBUG, "[%s] debug: %s", un, p);
		break;
	case 1:
		syslog(sf|LOG_INFO, "[%s] info: %s", un, p);
		break;
	case 2:
		syslog(sf|LOG_NOTICE, "[%s] notice: %s", un, p);
		break;
	case 3:
		syslog(sf|LOG_WARNING, "[%s] warn: %s", un, p);
		break;
	case 4:
		syslog(sf|LOG_ERR, "[%s] error: %s", un, p);
		break;
	case 5:
		syslog(sf|LOG_CRIT, "[%s] CRITICAL: %s", un, p);
		break;
	}
}

static void xlog(int severity, const char *fmt, ... )
{
	va_list ap;
	
	va_start(ap, fmt);
	log(severity, fmt, ap);
	va_end(ap);
}

static void elog(int severity, const char *fmt, va_list ap)
{
	int eno;
	char *p;
	
	eno = errno;
	p = xvprintf(fmt, ap);
	xlog(severity, "%s [%d: %s]", p, eno, strerror(eno));
}

void die(const char *fmt, ... )
{
	va_list ap;
	
	va_start(ap, fmt);
	elog(4, fmt, ap);
	va_end(ap);
	exit(1);
}

void warn(const char *fmt, ... )
{
	va_list ap;
	
	va_start(ap, fmt);
	log(3, fmt, ap);
	va_end(ap);
}

void ewarn(const char *fmt, ... )
{
	va_list ap;
	
	va_start(ap, fmt);
	elog(3, fmt, ap);
	va_end(ap);
}

void notice(const char *fmt, ... )
{
	va_list ap;
	
	va_start(ap, fmt);
	log(2, fmt, ap);
	va_end(ap);
}

void enotice(const char *fmt, ... )
{
	va_list ap;
	
	va_start(ap, fmt);
	elog(2, fmt, ap);
	va_end(ap);
}

void info(const char *fmt, ... )
{
	va_list ap;
	
	va_start(ap, fmt);
	log(1, fmt, ap);
	va_end(ap);
}

void einfo(const char *fmt, ... )
{
	va_list ap;
	
	va_start(ap, fmt);
	elog(1, fmt, ap);
	va_end(ap);
}

void debug(const char *fmt, ... )
{
	va_list ap;
	
	va_start(ap, fmt);
	log(0, fmt, ap);
	va_end(ap);
}

void error(const char *fmt, ... )
{
	va_list ap;
	
	va_start(ap, fmt);
	log(4, fmt, ap);
	va_end(ap);
}

void log_init()
{
	struct conf_map fac[] = {
		{"daemon", LOG_DAEMON},	/* default */
		{"ftp", LOG_FTP},
		{"user", LOG_USER},
		{"local0", LOG_LOCAL0},
		{"local1", LOG_LOCAL1},
		{"local2", LOG_LOCAL2},
		{"local3", LOG_LOCAL3},
		{"local4", LOG_LOCAL4},
		{"local5", LOG_LOCAL5},
		{"local6", LOG_LOCAL6},
		{"local7", LOG_LOCAL7},
		{0, 0}
	};
	struct conf_map pri[] = {
		{"debug", 0},
		{"info", 1},
		{"notice", 2},	/* default */
		{"warn", 3},
		{"error", 4},
		{"critical", 5},
		{"crit", 5},
		{"none", 6},
		{0, 0}
	};

	log_level = conf_get_maped("LogLevel", log_level, pri);
	syslog_f = conf_get_maped("LogType", syslog_f, fac);
#ifdef BRAINDEAD
	syslog(0, 0);
#else
	openlog(conf_get_string("LogName", "cftpd"),
		conf_get_bool("LogPID", 0) ? LOG_PID : 0, syslog_f);
#endif
}


#ifdef BRAINDEAD
static void syslog(int dummy, const char *fmt, ... )
{
	static FILE *log = 0;
	va_list ap;
	char *s;
	
	(void)dummy;
	
	if (log == 0) {
		log = fopen(ETC_DIR "/log", "a");
		if (log == 0) {
			perror(ETC_DIR "/log");
			exit(1);
		}
	}

	if (fmt == 0)
		return;
	
	va_start(ap, fmt);
	s = xvprintf(fmt, ap);
	va_end(ap);

	fprintf(log, "[] %s\n", s);
	fflush(log);
	printf("[] %s\n", s);
	xfree(s);
}
#endif

