/*
 * Copyright (c) y2k Micha Moskal <malekith@topnet.pl>
 *
 * This file is part of cryptoFTP - cftp protocol implementation.
 * CryptoFTP is free software. You can redistribute it and/or modify it under
 * the terms of the GNU General Public License (see COPYING.GPL for details).
 *
 * $Id: server_rsa.c,v 1.3 2000/04/23 16:39:03 malekith Exp $
 */

#include "config.h"
#include "server.h"
#include "rsa.h"

#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>


const char *rsa_private_key_name = RSA_PRIVATE_KEY_NAME;
const char *rsa_public_key_name = RSA_PUBLIC_KEY_NAME;

extern struct clib_state *ctx;

static void read_keys(char **pub, struct rsa_key *prv)
{
	/* can't use stdio here for security reasons -- after we setuid()ed
	 * user can access our address space so no bit of private key
	 * can left in memory */
	int fd;
	char buffer[5000];
	char *p, *e;
		
	memset(buffer, 0, sizeof(buffer));
	
	if ((fd = open(rsa_private_key_name, O_RDONLY)) == -1)
		die("can't open prvkey: '%s'", rsa_private_key_name);
	read(fd, buffer, sizeof(buffer) - 1);
	close(fd);
	
	if (strncmp(buffer, "cftp private key version 1.0\n", 29) != 0)
		die("it's not private key: %s", rsa_private_key_name);
		 
#define err(a) die("%s: key-file format error: " a , rsa_private_key_name)
	if ((p = strchr(buffer, '\n')) == 0) err("no magic nl (???)");
	p++;
	if ((p = strchr(p, '\n')) == 0) err("no bits");
	prv->bits = atoi(++p);
#define GET(n) \
	if ((e = strchr(p, '\n')) == 0) err("no " #n); \
	*e = 0; \
	mpz_init_set_str(prv->n, p, 16); \
	p = e + 1
	GET(n); GET(e); GET(d); /* GET(u); GET(p); GET(q);*/
	/* if (strchr(p, '\n')) err("2big"); */
#undef GET
#undef err
	
	memset(buffer, 0, sizeof(buffer));
	
	if ((fd = open(rsa_public_key_name, O_RDONLY)) == -1)
		die("can't open pubkey: '%s'", rsa_public_key_name);

	read(fd, buffer, sizeof(buffer) - 1);
	close(fd);
	*pub = strdup(buffer);

	memset(buffer, 0, sizeof(buffer));
}

struct cipher2 *using_cipher;
int using_key_len;

static void setup_pkey(struct pkt_key *pkey)
{
	const char *msg = "cipher OK";
	struct pkt_key_rsp *kr;
	struct cipher2 *c2;
	
	c2 = find_cipher(pkey->cipher_id);
	
	if (!c2)
		msg = "cipher not supported";
	else if (c2->disabled) {
		msg = c2->disabled;
		c2 = 0;
	}
	
	kr = (struct pkt_key_rsp *)xmalloc(2 + strlen(msg));
	
	kr->accepted = c2 ? 1 : 0;
	strcpy(kr->msg, msg);
	
	send_packet(ctx, PKT_KEY_RSP, (byte*)kr, strlen(msg) + 1);
	
	if (!c2) {
		sleep(1); /* wait for packet to reach client */
		die("cipher [%s] error: %s", pkey->cipher_id, msg);
	}
	
	ctx->u->e = c2->creator();
	ctx->u->d = c2->creator();

	ctx->u->e->init(ctx->u->e, pkey->server_key, pkey->used_key_len);
	ctx->u->d->init(ctx->u->d, pkey->client_key, pkey->used_key_len);
	ctx->u->encrypting = 1;
	
	using_cipher = c2;
	using_key_len = pkey->used_key_len;
	
	memset(pkey, 0, sizeof(*pkey));
}

void key_rsa()
{
	char *pub;
	struct rsa_key prv;
	struct packet *p;
	mpz_t blow;
	struct pkt_key pkey;
	
	crypto_init();
	
	read_keys(&pub, &prv);
	send_packet(ctx, PKT_RSAKEY, pub, strlen(pub));
	xfree0(pub, strlen(pub));
	p = recv_packet(ctx);
	if (p == 0)
		die("no packet (waiting for RSA key resp)");
	if (p->id != PKT_KEY)	
		die("PKT_RSAKEY response is not PKT_KEY [%d: %s]",
			p->id, p->data);
	mpz_init_set_str(blow, p->data, 16);
	free_packet(p);
	rsa_decrypt(blow, blow, &prv);
	rsa_destroy_key(&prv);
	xmpz_to_byte(blow, (byte*)&pkey, sizeof(pkey));
	mpz_clear(blow);
	
	setup_pkey(&pkey);
}
